/* --------------------------------------------------------------  */
/* (C)Copyright 2006,2008,                                         */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */
//
// Example:     2 - dacs_c Cell
// Language:    IBM XL C
// Core:        PPU
// File:        array_ppu.c
// Description: Function to communicate with the SPUs using
//              DaCS C APIs, called from main_ppu Fortran program
//

#include <stdio.h>
#include <errno.h>
#include <dacs.h>
#include "init_cb.h"
#include "dacs_chk.h"

// allocate control blocks, to correspond to one SPE
control_block cb1[MAX_UNIT_COUNT] __attribute__ ((aligned(128)));

// spe program to load
extern dacs_program_handle_t main_spu;

// error handler reference
uint32_t local_error_handler(dacs_error_t);


//
// PPU function to communicate to SPUs using DaCS APIs.
// Invoked from Fortran main_ppu program to distribute array elements
// to SPUs for processing.
//
void update_array_(float *carr, int n)
{
  int  i;
  int unit_size, unit_count, unit_buffer_size;

  DACS_ERR_T rc;
  int32_t status;
  uint32_t count_spes, num_rsvd_child;
  de_id_t rsvd_child_des[MAX_UNIT_COUNT];
  dacs_process_id_t rsvd_child_pids[MAX_UNIT_COUNT];
  dacs_remote_mem_t carr_remote_mem;
  dacs_wid_t ppu_wid[MAX_UNIT_COUNT];


  rc = dacs_init(0);
  ERRCHK("PPE: dacs_init", rc);

  rc = dacs_errhandler_reg(local_error_handler, 0);
  ERRCHK("PPE: dacs_errhandler_reg", rc);


  // get available spes count
  count_spes = MAX_UNIT_COUNT;
  rc = dacs_reserve_children(DACS_DE_SPE, &count_spes, rsvd_child_des);
  ERRCHK("PPE: dacs_reserve_children", rc);
  if (count_spes < 1) {
    fprintf(stderr, "Sample needs at least 1 SPE tor run.\n");
    exit(1);
  }

  //each spe processes an equal number of elements
  //element count must be a multiple of 8
  //for debug and analysis constants in init_ch.h can be modified to use fewer spes
  unit_size = (count_spes > 7)
              ? 8
              : (count_spes > 3)
                 ? 16
                 : (count_spes >1)
                   ? 32
                   : 64;
  //element count cannot be less than minimun in init_cb.h
  unit_size = (unit_size < MIN_UNIT_SIZE) ? MIN_UNIT_SIZE : unit_size;
  unit_count = DATA_BUFFER_ENTRIES/unit_size; //number of spes used in  process group
  unit_buffer_size = unit_size*sizeof(float); //size of buffer for a group
  num_rsvd_child = unit_count;

  //release extra children
  if (count_spes > num_rsvd_child) {
    rc = dacs_release_de_list(count_spes-num_rsvd_child, &rsvd_child_des[num_rsvd_child]);
  }
  PRINTF("array element count[%d] min spu element count[%d] spu element count[%d]\n", n, MIN_UNIT_SIZE,unit_size);
  PRINTF("max unit count[%d] spus available[%d] spus reserved[%d]\n", MAX_UNIT_COUNT, count_spes, unit_count);

  //create memory to be shared with spes
  rc = dacs_remote_mem_create(carr, DATA_BUFFER_SIZE,DACS_READ_WRITE, &carr_remote_mem);
  ERRCHK("PPE: dacs_remote_mem_create", rc);


  // initialize the control block for each unit of work
  for (i=0; i<unit_count; i++) {
    cb1[i].offset = i*unit_buffer_size;
    cb1[i].chunk_size = unit_buffer_size;
    cb1[i].index = i;
  }

  // allocate SPE tasks
  for (i=0; i<unit_count; i++) {
    //start the spe
    rc = dacs_de_start(rsvd_child_des[i], &main_spu, NULL, NULL, DACS_PROC_EMBEDDED,
                       &rsvd_child_pids[i]);
    ERRCHK("PPE: dacs_de_start", rc);

    //initialize a dacs wait identifier for use by dacs_send
    rc = dacs_wid_reserve(&ppu_wid[i]);
    ERRCHK("PPE: dacs_wid_reserve", rc);

    //send the cb
    rc = dacs_send(&cb1[i], sizeof(cb1[i]),rsvd_child_des[i], rsvd_child_pids[i],STREAM_ID,
                   ppu_wid[i], DACS_BYTE_SWAP_DISABLE);
    ERRCHK("PPE: dacs_send", rc);
  }

  //wait for each send to complete then share the memory
  for (i=0; i<unit_count; i++) {
    rc = dacs_wait(ppu_wid[i]);
    ERRCHK("PPE: dacs_wait", rc);


    rc = dacs_remote_mem_share(rsvd_child_des[i], rsvd_child_pids[i], carr_remote_mem);
    ERRCHK("PPE: dacs_remote_mem_share", rc);

  }

  PRINTF("SPEs 0 - %d are now computing...\n",unit_count);

  // wait for SPEs to all finish
  for (i=0; i<unit_count; ++i) {
    int index = cb1[i].index;
    PRINTF("ppu unit.%i waiting on completion\n",index);
    rc = dacs_de_wait(rsvd_child_des[i], rsvd_child_pids[i], &status);
    if (rc != DACS_STS_PROC_FINISHED) {
       PRINTF("dacs_de_wait rc[%d] - %s\n", rc, dacs_strerror(rc));
    }

  }

  // cleanup resources
  for (i=0; i<unit_count; ++i) {
    rc = dacs_wid_release(&ppu_wid[i]);
    ERRCHK("PPE: dacs_wid_release", rc);
  }

  rc = dacs_remote_mem_destroy(&carr_remote_mem);
  ERRCHK("PPE: dacs_remote_mem_destroy", rc);

  PRINTF("ppu exiting\n");
  dacs_exit();
}

uint32_t
local_error_handler(dacs_error_t error)
{
    de_id_t de;
    dacs_process_id_t pid;
    const char *str;
    uint32_t code;

    // Use the dacs error interfaces to determine the error details
    dacs_error_code(error, &code);
    dacs_error_str(error, &str);
    dacs_error_de(error, &de);
    dacs_error_pid(error, &pid);
    PRINTF("PPE: a %s (%d) error on DE/PID 0x%x/0x%lx\n",
            str, code, de, pid);

    return 0;
}

