/* --------------------------------------------------------------  */
/* (C)Copyright 2006,2008,                                         */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */
//
// Example:     1 - spe2_c Cell
// Language:    IBM XL C
// Core:        PPU
// File:        array_ppu.c
// Description: Function to communicate with the SPUs using
//              libspe2 APIs, called from main_ppu Fortran program
//

#include <sched.h>
#include <libspe2.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "init_cb.h"

typedef struct ppu_pthread_data {
  spe_context_ptr_t speid;
  pthread_t pthread;
  void *argp;
} ppu_pthread_data_t;

// thread used to run the spes
void *ppu_pthread_function(void *arg) {
  ppu_pthread_data_t *datap = (ppu_pthread_data_t *)arg;
  int rc;
  unsigned int entry = SPE_DEFAULT_ENTRY;
  if ((rc=spe_context_run(datap->speid, &entry, 0, datap->argp, NULL, NULL)) < 0) {
    fprintf(stderr, "Failed spe_context_run(rc=%d, errno=%d, strerror=%s)\n",
    rc, errno, strerror(errno));
    exit (1);
  }
  pthread_exit(NULL);
}

// allocate control blocks, to correspond to one SPE
control_block cb1[UNIT_COUNT] __attribute__ ((aligned(128)));

// spe program to load
extern spe_program_handle_t main_spu;

// thread data for each unit of work
ppu_pthread_data_t datas[UNIT_COUNT];

//
// PPU function to communicate to SPUs using libspe2 APIs.
// Invoked from Fortran main_ppu program to distribute array elements
// to SPUs for processing.
//
void update_array_(float *carr, int n)
{
  int  i, j, count_spes, rc;

  // get available spes count
  count_spes = spe_cpu_info_get(SPE_COUNT_PHYSICAL_SPES, -1);
  if (count_spes>8) {
    count_spes=8;
  }
  if (count_spes < 1) {
    fprintf(stderr, "Sample needs at least 1 SPE tor run.\n");
    exit(1);
  }

  printf("n=%d tasks=%d count_spes=%d \n", n, UNIT_COUNT, count_spes);

  // load the address into the control block for each unit of work
  for (i=0; i<UNIT_COUNT; i++) {
    cb1[i].addr.p = &carr[i*UNIT_SIZE];
    cb1[i].chunk_size = UNIT_BUFFER_SIZE;
  }

  // process units of work in groups by spe count
  for (j=0;j<UNIT_COUNT;j+=count_spes) {
    // group count
    int imax = (j+count_spes <= UNIT_COUNT) ? count_spes : UNIT_COUNT-j;
    // initialize and start SPE within group
    for (i=0; i<imax; i++) {

      // create the SPE context
      if ((datas[j+i].speid = spe_context_create(0, NULL)) == NULL) {
        fprintf(stderr, "FAILED: spe_context_create(errno=%d strerror=%s)\n",
                errno, strerror(errno));
        exit(3+i);
      }

      // load the SPE program into the SPE context
      if (spe_program_load(datas[j+i].speid, &main_spu) != 0) {
        fprintf(stderr, "FAILED: spe_program_load(errno=%d strerror=%s)\n",
                errno, strerror(errno));
        exit(3+i);
      }

      datas[j+i].argp = &cb1[j+i]; //cb is passed as argp
      printf("sending cb[%i].addr[%p] chunksize[%i]\n",i,cb1[i].addr.p,cb1[i].chunk_size);
      // Create thread
      if ((rc=pthread_create(&datas[j+i].pthread, NULL, &ppu_pthread_function,
                             &datas[j+i])) != 0) {
        fprintf(stderr, "Failed pthread_create(errno=%d strerror=%s)\n",
                errno, strerror(errno));
        exit(3+i);
      }

    }

    printf("SPEs 0 - %d are now computing...\n",imax);

    // wait for SPEs in group to all finish
    for (i=0; i<imax; ++i) {
      // Join thread
      if ((rc = pthread_join(datas[i+j].pthread, NULL)) != 0) {
        fprintf(stderr, "Failed pthread_join(rc=%d, errno=%d strerror=%s)\n",
                rc, errno, strerror(errno));
        exit(1);
      }

      // Destroy context
      if ((rc = spe_context_destroy(datas[i+j].speid)) != 0) {
        fprintf(stderr, "Failed spe_context_destroy(rc=%d, errno=%d strerror=%s)\n", rc, errno, strerror(errno));
        exit(1);
      }
    }
  }

  // issue a sync to be safe
  __asm__ __volatile__ ("sync" : : : "memory");

  printf("SPEs done.\n");
}
